"""
-----------
drfseriperm
-----------

Library for Django REST framework which provides permissions-based serialization

:copyright: (c) 2025-present DouleLove
:license: MIT, see LICENSE file for more details.
"""

__title__ = 'drfseriperm'
__author__ = 'DouleLove'
__license__ = 'MIT'
__copyright__ = 'Copyright 2025-present DouleLove'
__version__ = '1.0.2'

from .ffp import *
from .missing import *
from .mixins import *
